@@ -1,9 +1,9 @@
 /*
- * Copyright © 2013-2018 Inria.  All rights reserved.
+ * Copyright © 2013-2019 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
-#include <private/private.h>
+#include <private/autogen/config.h>
 #include <hwloc.h>
 #include <hwloc/diff.h>
 
@@ -22,19 +22,22 @@ int main(int argc, char *argv[])
 {
 	hwloc_topology_t topo1, topo2;
 	hwloc_topology_diff_t firstdiff = NULL, diff;
-	unsigned long flags = HWLOC_TOPOLOGY_FLAG_WHOLE_SYSTEM | HWLOC_TOPOLOGY_FLAG_WHOLE_IO | HWLOC_TOPOLOGY_FLAG_ICACHES;
+	unsigned long flags = HWLOC_TOPOLOGY_FLAG_WHOLE_SYSTEM;
 	char *callname, *input1, *input2, *output, *outputname, *refname = NULL;
 	char *xmlbuffer;
 	int xmlbuflen;
 	unsigned i, j;
 	int err;
 
-	putenv((char *) "HWLOC_XML_VERBOSE=1");
 	callname = argv[0];
 	/* skip argv[0], handle options */
 	argc--;
 	argv++;
 
+	hwloc_utils_check_api_version(callname);
+
+	putenv((char *) "HWLOC_XML_VERBOSE=1");
+
 	while (argc && *argv[0] == '-') {
 		if (!strcmp (argv[0], "--refname")) {
 			if (argc < 2) {
@@ -75,22 +78,32 @@ int main(int argc, char *argv[])
 	}
 
 	hwloc_topology_init(&topo1);
+	hwloc_topology_set_all_types_filter(topo1, HWLOC_TYPE_FILTER_KEEP_ALL);
 	hwloc_topology_set_flags(topo1, flags);
 	err = hwloc_topology_set_xml(topo1, input1);
 	if (err < 0) {
 		fprintf(stderr, "Failed to load 1st XML topology %s\n", input1);
-		goto out;
+		goto out_with_topo1;
+	}
+	err = hwloc_topology_load(topo1);
+	if (err < 0) {
+		fprintf(stderr, "Failed to load 1st topology %s\n", input1);
+		goto out_with_topo1;
 	}
-	hwloc_topology_load(topo1);
 
 	hwloc_topology_init(&topo2);
+	hwloc_topology_set_all_types_filter(topo2, HWLOC_TYPE_FILTER_KEEP_ALL);
 	hwloc_topology_set_flags(topo2, flags);
 	err = hwloc_topology_set_xml(topo2, input2);
 	if (err < 0) {
 		fprintf(stderr, "Failed to load 2nd XML topology %s\n", input2);
-		goto out_with_topo1;
+		goto out_with_topo2;
+	}
+	err = hwloc_topology_load(topo2);
+	if (err < 0) {
+		fprintf(stderr, "Failed to load 2nd topology %s\n", input2);
+		goto out_with_topo2;
 	}
-	hwloc_topology_load(topo2);
 
 	if (!refname) {
 		refname = strrchr(input1, '/');
@@ -124,9 +137,9 @@ int main(int argc, char *argv[])
 	}
 	if (!j) {
 		if (output) {
-			err = hwloc_topology_diff_export_xml(topo1, firstdiff, refname, output);
+			err = hwloc_topology_diff_export_xml(firstdiff, refname, output);
 		} else {
-			err = hwloc_topology_diff_export_xmlbuffer(topo1, firstdiff, refname, &xmlbuffer, &xmlbuflen);
+			err = hwloc_topology_diff_export_xmlbuffer(firstdiff, refname, &xmlbuffer, &xmlbuflen);
 			if (!err) {
 				printf("%s\n", xmlbuffer);
 				hwloc_free_xmlbuffer(topo1, xmlbuffer);
@@ -136,7 +149,7 @@ int main(int argc, char *argv[])
 			fprintf(stderr, "Failed to export topology diff %s\n", output);
 	}
 
-	hwloc_topology_diff_destroy(topo1, firstdiff);
+	hwloc_topology_diff_destroy(firstdiff);
 
 	hwloc_topology_destroy(topo2);
 	hwloc_topology_destroy(topo1);
@@ -150,6 +163,5 @@ out_with_topo2:
 	hwloc_topology_destroy(topo2);
 out_with_topo1:
 	hwloc_topology_destroy(topo1);
-out:
 	exit(EXIT_FAILURE);
 }
