@@ -1,6 +1,6 @@
 /*
  * Copyright © 2009 CNRS
- * Copyright © 2009-2017 Inria.  All rights reserved.
+ * Copyright © 2009-2018 Inria.  All rights reserved.
  * Copyright © 2009-2010, 2012 Université Bordeaux
  * Copyright © 2011 Cisco Systems, Inc.  All rights reserved.
  * See COPYING in top-level directory.
@@ -24,28 +24,26 @@
 /* windows back-end.  */
 
 static struct color {
-  int r, g, b;
+  struct lstopo_color color;
   HGDIOBJ brush;
 } *colors;
 
 struct lstopo_windows_output {
-  struct lstopo_output loutput; /* must be at the beginning */
-  int drawing;
+  struct lstopo_output *loutput;
   PAINTSTRUCT ps;
   HWND toplevel;
-  unsigned max_x;
-  unsigned max_y;
 };
 
 static int numcolors;
 
 static HGDIOBJ
-rgb_to_brush(int r, int g, int b)
+lcolor_to_brush(const struct lstopo_color *lcolor)
 {
+  int r = lcolor->r, g = lcolor->g, b = lcolor->b;
   int i;
 
   for (i = 0; i < numcolors; i++)
-    if (colors[i].r == r && colors[i].g == g && colors[i].b == b)
+    if (colors[i].color.r == r && colors[i].color.g == g && colors[i].color.b == b)
       return colors[i].brush;
 
   fprintf(stderr, "color #%02x%02x%02x not declared\n", r, g, b);
@@ -64,11 +62,12 @@ static unsigned int the_fontsize, the_gridsize;
 static float the_scale;
 
 static void
-windows_box(void *output, int r, int g, int b, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned width, unsigned y, unsigned height);
+windows_box(struct lstopo_output *loutput, const struct lstopo_color *lcolor, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned width, unsigned y, unsigned height);
 
 static LRESULT CALLBACK
 WndProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
 {
+  struct lstopo_output *loutput = the_output.loutput;
   int redraw = 0;
   switch (message) {
     case WM_CHAR:  {
@@ -104,15 +103,18 @@ WndProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
 
     case WM_PAINT: {
       HFONT font;
+      struct lstopo_color white = {0xff, 0xff, 0xff};
       BeginPaint(hwnd, &the_output.ps);
-      font = CreateFont(fontsize, 0, 0, 0, 0, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, NULL);
+      font = CreateFont(loutput->fontsize, 0, 0, 0, 0, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, NULL);
       SelectObject(the_output.ps.hdc, (HGDIOBJ) font);
-      windows_box(&the_output, 0xff, 0xff, 0xff, 0, 0, win_width, 0, win_height);
-      the_output.max_x = 0;
-      the_output.max_y = 0;
-      output_draw(&the_output.loutput);
-      the_width = the_output.max_x;
-      the_height = the_output.max_y;
+      SetBkMode(the_output.ps.hdc, TRANSPARENT);
+      windows_box(loutput, &white, 0, 0, win_width, 0, win_height);
+      loutput->drawing = LSTOPO_DRAWING_PREPARE;
+      output_draw(loutput);
+      the_width = loutput->width;
+      the_height = loutput->height;
+      loutput->drawing = LSTOPO_DRAWING_DRAW;
+      output_draw(loutput);
       DeleteObject(font);
       EndPaint(hwnd, &the_output.ps);
       break;
@@ -226,24 +228,108 @@ WndProc(HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam)
       x_delta = 0;
     if (y_delta < 0)
       y_delta = 0;
-    fontsize = (unsigned)(the_fontsize * the_scale);
-    gridsize = (unsigned)(the_gridsize * the_scale);
+    loutput->fontsize = (unsigned)(the_fontsize * the_scale);
+    loutput->gridsize = (unsigned)(the_gridsize * the_scale);
     RedrawWindow(hwnd, NULL, NULL, RDW_INVALIDATE);
   }
   return DefWindowProc(hwnd, message, wparam, lparam);
 }
 
 static void
-windows_init(void *output)
+windows_declare_color(struct lstopo_output *loutput __hwloc_attribute_unused, const struct lstopo_color *lcolor)
+{
+  int r = lcolor->r, g = lcolor->g, b = lcolor->b;
+  HBRUSH brush;
+  COLORREF color;
+  struct color *tmp;
+
+  color = RGB(r, g, b);
+  brush = CreateSolidBrush(color);
+  if (!brush) {
+    fprintf(stderr,"Could not allocate color %02x%02x%02x\n", r, g, b);
+    exit(EXIT_FAILURE);
+  }
+
+  tmp = realloc(colors, sizeof(*colors) * (numcolors + 1));
+  if (!tmp) {
+    fprintf(stderr, "Failed to realloc the colors array\n");
+    return;
+  }
+  colors = tmp;
+  colors[numcolors].color = *lcolor;
+  colors[numcolors].brush = (HGDIOBJ) brush;
+  numcolors++;
+}
+
+static void
+windows_box(struct lstopo_output *loutput, const struct lstopo_color *lcolor, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned width, unsigned y, unsigned height)
+{
+  struct lstopo_windows_output *woutput = loutput->backend_data;
+  PAINTSTRUCT *ps = &woutput->ps;
+  int r = lcolor->r, g = lcolor->g, b = lcolor->b;
+
+  SelectObject(ps->hdc, lcolor_to_brush(lcolor));
+  SetBkColor(ps->hdc, RGB(r, g, b));
+  Rectangle(ps->hdc, x - x_delta, y - y_delta, x + width - x_delta, y + height - y_delta);
+}
+
+static void
+windows_line(struct lstopo_output *loutput, const struct lstopo_color *lcolor, unsigned depth __hwloc_attribute_unused, unsigned x1, unsigned y1, unsigned x2, unsigned y2)
+{
+  struct lstopo_windows_output *woutput = loutput->backend_data;
+  PAINTSTRUCT *ps = &woutput->ps;
+
+  SelectObject(ps->hdc, lcolor_to_brush(lcolor));
+  MoveToEx(ps->hdc, x1 - x_delta, y1 - y_delta, NULL);
+  LineTo(ps->hdc, x2 - x_delta, y2 - y_delta);
+}
+
+static void
+windows_text(struct lstopo_output *loutput, const struct lstopo_color *lcolor, int size __hwloc_attribute_unused, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned y, const char *text)
+{
+  struct lstopo_windows_output *woutput = loutput->backend_data;
+  PAINTSTRUCT *ps = &woutput->ps;
+  int r = lcolor->r, g = lcolor->g, b = lcolor->b;
+
+  SetTextColor(ps->hdc, RGB(r, g, b));
+  TextOut(ps->hdc, x - x_delta, y - y_delta, text, (int)strlen(text));
+}
+
+static void
+windows_textsize(struct lstopo_output *loutput, const char *text, unsigned textlength, unsigned fontsize __hwloc_attribute_unused, unsigned *width)
+{
+  struct lstopo_windows_output *woutput = loutput->backend_data;
+  PAINTSTRUCT *ps = &woutput->ps;
+  SIZE size;
+
+  GetTextExtentPoint32(ps->hdc, text, textlength, &size);
+  *width = size.cx;
+}
+
+struct draw_methods windows_draw_methods = {
+  windows_declare_color,
+  windows_box,
+  windows_line,
+  windows_text,
+  windows_textsize,
+};
+
+int
+output_windows (struct lstopo_output *loutput, const char *dummy __hwloc_attribute_unused)
 {
-  struct lstopo_windows_output *woutput = output;
   WNDCLASS wndclass;
   HWND toplevel, faketoplevel;
   unsigned width, height;
   HFONT font;
+  MSG msg;
+
+  memset(&the_output, 0, sizeof(the_output));
+  the_output.loutput = loutput;
+  loutput->methods = &windows_draw_methods;
+  loutput->backend_data = &the_output;
 
   /* make sure WM_DESTROY on the faketoplevel won't kill the program */
-  woutput->toplevel = NULL;
+  the_output.toplevel = NULL;
 
   /* create the toplevel window, with random size for now */
   memset(&wndclass, 0, sizeof(wndclass));
@@ -255,25 +341,23 @@ windows_init(void *output)
 
   RegisterClass(&wndclass);
 
-  /* compute the maximal needed size, this may require the toplevel window in the future */
-  woutput->max_x = 0;
-  woutput->max_y = 0;
-  woutput->drawing = 0;
+  /* recurse once for preparing sizes and positions using a fake top level window */
+  loutput->drawing = LSTOPO_DRAWING_PREPARE;
   faketoplevel = CreateWindow("lstopo", "lstopo", WS_OVERLAPPEDWINDOW,
 			      CW_USEDEFAULT, CW_USEDEFAULT,
 			      10, 10, NULL, NULL, NULL, NULL);
-  BeginPaint(faketoplevel, &woutput->ps);
-  font = CreateFont(fontsize, 0, 0, 0, 0, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, NULL);
-  SelectObject(woutput->ps.hdc, (HGDIOBJ) font);
-  output_draw(&woutput->loutput);
+  BeginPaint(faketoplevel, &the_output.ps);
+  font = CreateFont(loutput->fontsize, 0, 0, 0, 0, FALSE, FALSE, FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH, NULL);
+  SelectObject(the_output.ps.hdc, (HGDIOBJ) font);
+  output_draw(loutput);
   DeleteObject(font);
-  EndPaint(faketoplevel, &woutput->ps);
+  EndPaint(faketoplevel, &the_output.ps);
   DestroyWindow(faketoplevel);
-  woutput->drawing = 1;
+  loutput->drawing = LSTOPO_DRAWING_DRAW;
 
   /* now create the actual toplevel with the sizes */
-  width = woutput->max_x;
-  height = woutput->max_y;
+  width = loutput->width;
+  height = loutput->height;
 
   win_width = width + 2*GetSystemMetrics(SM_CXSIZEFRAME);
   win_height = height + 2*GetSystemMetrics(SM_CYSIZEFRAME) + GetSystemMetrics(SM_CYCAPTION);
@@ -287,15 +371,15 @@ windows_init(void *output)
   toplevel = CreateWindow("lstopo", "lstopo", WS_OVERLAPPEDWINDOW,
 		  CW_USEDEFAULT, CW_USEDEFAULT,
 		  win_width, win_height, NULL, NULL, NULL, NULL);
-  woutput->toplevel = toplevel;
+  the_output.toplevel = toplevel;
 
   the_width = width;
   the_height = height;
 
   the_scale = 1.0f;
 
-  the_fontsize = fontsize;
-  the_gridsize = gridsize;
+  the_fontsize = loutput->fontsize;
+  the_gridsize = loutput->gridsize;
 
   /* and display the window */
   ShowWindow(toplevel, SW_SHOWDEFAULT);
@@ -311,131 +395,17 @@ windows_init(void *output)
   printf(" Scroll to the bottom-right corner . End\n");
   printf(" Exit .............................. q Q Esc\n");
   printf("\n\n");
-}
+  fflush(stdout);
 
-static void
-windows_declare_color(void *output, int r, int g, int b)
-{
-  struct lstopo_windows_output *woutput = output;
-  HBRUSH brush;
-  COLORREF color;
-  struct color *tmp;
+  /* ready */
+  declare_colors(loutput);
+  lstopo_prepare_custom_styles(loutput);
 
-  if (!woutput->drawing)
-    return;
-
-  color = RGB(r, g, b);
-  brush = CreateSolidBrush(color);
-  if (!brush) {
-    fprintf(stderr,"Could not allocate color %02x%02x%02x\n", r, g, b);
-    exit(EXIT_FAILURE);
-  }
-
-  tmp = realloc(colors, sizeof(*colors) * (numcolors + 1));
-  if (!tmp) {
-    fprintf(stderr, "Failed to realloc the colors array\n");
-    return;
-  }
-  colors = tmp;
-  colors[numcolors].r = r;
-  colors[numcolors].g = g;
-  colors[numcolors].b = b;
-  colors[numcolors].brush = (HGDIOBJ) brush;
-  numcolors++;
-}
-
-static void
-windows_box(void *output, int r, int g, int b, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned width, unsigned y, unsigned height)
-{
-  struct lstopo_windows_output *woutput = output;
-  PAINTSTRUCT *ps = &woutput->ps;
-
-  if (x > woutput->max_x)
-    woutput->max_x = x;
-  if (x+width > woutput->max_x)
-    woutput->max_x = x + width;
-  if (y > woutput->max_y)
-    woutput->max_y = y;
-  if (y + height > woutput->max_y)
-    woutput->max_y = y + height;
-
-  if (!woutput->drawing)
-    return;
-
-  SelectObject(ps->hdc, rgb_to_brush(r, g, b));
-  SetBkColor(ps->hdc, RGB(r, g, b));
-  Rectangle(ps->hdc, x - x_delta, y - y_delta, x + width - x_delta, y + height - y_delta);
-}
-
-static void
-windows_line(void *output, int r, int g, int b, unsigned depth __hwloc_attribute_unused, unsigned x1, unsigned y1, unsigned x2, unsigned y2)
-{
-  struct lstopo_windows_output *woutput = output;
-  PAINTSTRUCT *ps = &woutput->ps;
-
-  if (x1 > woutput->max_x)
-    woutput->max_x = x1;
-  if (x2 > woutput->max_x)
-    woutput->max_x = x2;
-  if (y1 > woutput->max_y)
-    woutput->max_y = y1;
-  if (y2 > woutput->max_y)
-    woutput->max_y = y2;
-
-  if (!woutput->drawing)
-    return;
-
-  SelectObject(ps->hdc, rgb_to_brush(r, g, b));
-  MoveToEx(ps->hdc, x1 - x_delta, y1 - y_delta, NULL);
-  LineTo(ps->hdc, x2 - x_delta, y2 - y_delta);
-}
-
-static void
-windows_text(void *output, int r, int g, int b, unsigned depth __hwloc_attribute_unused, unsigned x, unsigned y, const char *text)
-{
-  struct lstopo_windows_output *woutput = output;
-  PAINTSTRUCT *ps = &woutput->ps;
-
-  if (!woutput->drawing)
-    return;
-
-  SetTextColor(ps->hdc, RGB(r, g, b));
-  TextOut(ps->hdc, x - x_delta, y - y_delta, text, (int)strlen(text));
-}
-
-static void
-windows_textsize(void *output, const char *text, unsigned textlength, unsigned *width)
-{
-  struct lstopo_windows_output *woutput = output;
-  PAINTSTRUCT *ps = &woutput->ps;
-  SIZE size;
-
-  GetTextExtentPoint32(ps->hdc, text, textlength, &size);
-  *width = size.cx;
-}
-
-struct draw_methods windows_draw_methods = {
-  windows_init,
-  windows_declare_color,
-  windows_box,
-  windows_line,
-  windows_text,
-  windows_textsize,
-};
-
-void
-output_windows (struct lstopo_output *loutput, const char *filename __hwloc_attribute_unused)
-{
-  MSG msg;
-
-  memset(&the_output, 0, sizeof(the_output));
-  memcpy(&the_output.loutput, loutput, sizeof(*loutput));
-  the_output.loutput.methods = &windows_draw_methods;
-
-  output_draw_start(&the_output.loutput);
   UpdateWindow(the_output.toplevel);
   while (!finish && GetMessage(&msg, NULL, 0, 0)) {
     TranslateMessage(&msg);
     DispatchMessage(&msg);
   }
+
+  return 0;
 }
